﻿using System.Collections.Generic;
using UnityEngine;


/// <summary>
/// プレーヤーキャラクタAnimatorの処理。
/// GameCharStateBehaviourと連動して現在の状態の記録もする。
/// </summary>
public class CharacterAnimatorController : MonoBehaviour
{
    [SerializeField] private GameObject attackCollider;
    public bool enableActionTransition = true;

    private Animator animator;
    private Dictionary<GameAnimType, int> enterCounts = new Dictionary<GameAnimType, int>();

    private TransitionId transitionId
    {
        get { return (TransitionId) animator.GetInteger("Action"); }
        set
        {
            if (enableActionTransition) {
                animator.SetInteger("Action", (int) value);
            }
        }
    }

    private enum TransitionId
    {
        None = 0,
        Locomotion = 1,
        Attack = 2,
        Damage = 3,
    }

    void Awake()
    {
        animator = GetComponent<Animator>();
    }

    void Start()
    {
        if (attackCollider != null) {
            attackCollider.SetActive(false);
        }
    }

    public bool IsState(GameAnimType state)
    {
        return enterCounts.ContainsKey(state) && enterCounts[state] > 0;
    }

    public void OnStateEnter(GameAnimType state)
    {
        if (enterCounts.ContainsKey(state) == false) {
            enterCounts[state] = 0;
        }
        enterCounts[state]++;
    }

    public void OnStateExit(GameAnimType state)
    {
        if (enterCounts.ContainsKey(state)) {
            enterCounts[state]--;
        }
    }

    public void OnStartTime(GameCharStateBehaviour sb, float now)
    {
        switch (sb.state) {
            case GameAnimType.AttackMotion:
                attackCollider.SetActive(true);
                break;
            case GameAnimType.ResetActionParameter:
                transitionId = TransitionId.None;
                break;
            default:
                break;
        }
    }

    public void OnEndTime(GameCharStateBehaviour sb, float now)
    {
        if (sb.state == GameAnimType.AttackMotion) {
            attackCollider.SetActive(false);
        }
    }

    public void Reset()
    {
        transitionId = TransitionId.None;
        animator.Rebind();
    }

    public void SetLocomotion(bool b)
    {
        if (CanTransitTo(TransitionId.Locomotion)) {
            transitionId = b ? TransitionId.Locomotion : TransitionId.None;
        } else {
            Debug.LogError("cannot transit : " + transitionId);
        }
    }

    public void Attack()
    {
        if (CanTransitTo(TransitionId.Attack)) {
            transitionId = TransitionId.Attack;
        }
    }

    public void GetDamage()
    {
        if (CanTransitTo(TransitionId.Damage)) {
            transitionId = TransitionId.Damage;
        }
    }

    private bool CanTransitTo(TransitionId tid)
    {
        return (int) tid >= (int) transitionId;
    }

    /// <summary>
    /// OnAnimatorMove() を定義したMonoBehaviourを持つキャラクタのAnimatorのapply root motionは無効になる。
    /// かわりに OnAnimatorMove()内で移動に関する処理をする。
    /// </summary>
    void OnAnimatorMove()
    {
        if (IsState(GameAnimType.ApplyRootMotion)) {
            animator.ApplyBuiltinRootMotion();
        }
    }
}
