﻿using System;
using UnityEngine;

/// <summary>
/// プレーヤーキャラクタのアニメーション状態。
/// </summary>
public enum GameAnimType
{
    // 移動へ遷移不能
    BlockLocomotion,
    // 攻撃中(キック)
    AttackMotion,
    // モーションの移動量をキャラクタの移動に反映
    ApplyRootMotion,
    // 
    ResetActionParameter,
}

/// <summary>
/// Animator State に属性(GameAnimType)を一つつける StateMachineBehaviour。
/// 設定した開始時間/終了時間に CharacterAnimatorController.OnStateEnd() / OnStateEnd() を呼ぶ。
/// </summary>
public class GameCharStateBehaviour : StateMachineBehaviour
{
    public GameAnimType state;
    [SerializeField] private float startTime = 0f;
    [SerializeField] private float endTime = 1f;

    private CharacterAnimatorController controller;

    private float prevFrameTime;

    override public void OnStateEnter(Animator animator, AnimatorStateInfo stateInfo, int layerIndex)
    {
        if (controller == null) {
            controller = animator.GetComponent<CharacterAnimatorController>();
        }
        if (controller != null) {
            controller.OnStateEnter(state);
        }
        prevFrameTime = -1f;
    }

    public override void OnStateUpdate(Animator animator, AnimatorStateInfo stateInfo, int layerIndex)
    {
        var now = stateInfo.normalizedTime;
        if (IsTime(startTime, now)) {
            OnStartTime(now);
        } else if (IsTime(endTime, now)) {
            OnEndTime(now);
        }
        prevFrameTime = now;
    }

    private void OnStartTime(float now)
    {
        if (controller != null) {
            controller.OnStartTime(this, now);
        }
    }

    private void OnEndTime(float now)
    {
        if (controller != null) {
            controller.OnEndTime(this, now);
        }
    }

    override public void OnStateExit(Animator animator, AnimatorStateInfo stateInfo, int layerIndex)
    {
        if (controller != null) {
            controller.OnStateExit(state);
        }
        var now = stateInfo.normalizedTime;
        if (prevFrameTime < endTime) {
            OnEndTime(now);
        }
    }

    private bool IsTime(float t, float now)
    {
        return prevFrameTime < t && t <= now;
    }
}
