﻿using UnityEngine;
using UnityEngine.Networking;
using UnityEngine.Events;
using System.Collections;

[System.Serializable]
public class GameTimeEvent : UnityEvent<float> { }


/// <summary>
/// ゲームコントローラー。
/// ゲーム時間の同期、ゲーム開始、ゲーム終了処理を行う。
/// </summary>
public class GameController : NetworkBehaviour
{
    public static GameController instance;

    [SerializeField] private const int timeLimit = 120;
    [SerializeField] private const int startCountDownTime = 3;
    [SerializeField] private Hud hud;
    [SerializeField] private bool useTestUI;

    public static bool isClockAvailable { get { return instance != null && (instance.isServer || instance.latency > 0f); } }
    public static bool isInPlay { get { return isClockAvailable && 0 < instance.remainingTime && instance.remainingTime < timeLimit; } }
    private float remainingTime { get { return (timeLimit + startCountDownTime) - (worldClock - serverStartTime); } }
    public float worldClock { get { return localClock + clockCorrectDelta; } }

    private const int MsgClientGameTimeMessageId = 1100;
    private const int MsgServerGameTimeMessageId = 1101;

    private float clientSendIntervalTime = 0f;
    private float latency;
    private float clientDeltaTime;
    private float clockCorrectDelta = 0f;
    private bool timeUpStarted = false;
    private float lastSendTime;

    private static float localClock
    {
        get
        { // DateTime の方を使うと、同一マシン上で複数起動しても同値になるので誤差や計算の確認に便利だが、h/m/s/msecで計算するので日を跨ぐと不具合を起こす。
            return Time.realtimeSinceStartup;
            //        var dt = System.DateTime.UtcNow;
            //        return dt.Hour * 3600 + dt.Minute * 60 + dt.Second + dt.Millisecond / 1000f;
        }
    }

    [SyncVar]
    public float serverStartTime;
    private bool transformNetworkGui = true;

    void Awake()
    {
        instance = this;
    }

    public override void OnStartServer()
    {
        serverStartTime = localClock;
        NetworkServer.RegisterHandler(MsgClientGameTimeMessageId, OnServerReceiveGameTime);
    }

    public override void OnStartClient()
    {
        hud.Init();
        var client = NetworkManager.singleton.client;
        if (client != null) {
            client.RegisterHandler(MsgServerGameTimeMessageId, OnClientReceiveGameTime);
        }
    }

    void Update()
    {
        //                Debug.Log("RTT: " + isServer + "/" + NetworkManager.singleton.client.GetRTT());
        if (isServer) {
            if (remainingTime <= 0f && !timeUpStarted) {
                timeUpStarted = true;
                StartCoroutine(TimeUp());
            }
        }
        if (isClient) { // クライアントだけの処理
            clientSendIntervalTime -= Time.unscaledDeltaTime;
            if (clientSendIntervalTime <= 0f) { // １秒おきに送信
                clientSendIntervalTime = 1f;
                //  msg = 送信データ。 lastSendTime = localClock = ローカル時刻
                lastSendTime = localClock; // 最後に送信したクライアント時刻を覚えておく
                var msg = new GameTimeClientMessage { sendTime = lastSendTime };
                NetworkManager.singleton.client.SendUnreliable(MsgClientGameTimeMessageId, msg);
                var me = NetworkClient.allClients;
            }
        }

        var countDownTime = remainingTime - timeLimit;
        hud.SetTime(Mathf.Clamp(remainingTime, 0f, timeLimit));
        hud.SetCountDown(Mathf.Max(0f, countDownTime + 1f));
    }


    private void OnServerReceiveGameTime(NetworkMessage netMsg)
    {
        var msg = netMsg.ReadMessage<GameTimeClientMessage>();
        var rtMsg = new GameTimeServerMessage { sendTime = msg.sendTime, serverTime = localClock };
        NetworkServer.SendToClient(netMsg.conn.connectionId, MsgServerGameTimeMessageId, rtMsg);
    }

    private void OnClientReceiveGameTime(NetworkMessage netMsg)
    {
        var msg = netMsg.ReadMessage<GameTimeServerMessage>();
        if (lastSendTime != msg.sendTime) {
            var now = localClock;
            latency = (now - msg.sendTime) * 0.5f;
            clientDeltaTime = msg.serverTime - now;
            clockCorrectDelta = clientDeltaTime + latency;
        }
        // now = localClock がどのクライアント(サーバー)でも同じ値のとき（全クライアント同一PC上＋localClockの get{} 実装修正）だけ有効な誤差確認。
        // var err = currentClock - now;
        // Debug.Log("同期誤差=" + err + " 現在時刻=" + now + " 補正時刻=" + currentClock + " レイテンシ=" + latency);
    }

    void OnGUI()
    {
        if (!useTestUI) {
            return;
        }
        GUILayout.Label("**** " + localClock);
        if (isServer) {
            GUILayout.Label("server: " + localClock);
        } else {
            GUILayout.Label("client: " + worldClock);
            GUILayout.Label("latency: " + latency);
        }
        if (transformNetworkGui) {
            if (GUILayout.Button("Reckoning:" + PainterNetworkTransform.enableReckoning)) {
                PainterNetworkTransform.enableReckoning = !PainterNetworkTransform.enableReckoning;
            }
            if (GUILayout.Button("Interpolation:" + PainterNetworkTransform.enableInterpolation)) {
                PainterNetworkTransform.enableInterpolation = !PainterNetworkTransform.enableInterpolation;
            }
        }

        var reset = false;
        if (GUILayout.Button("Reset statistics")) {
            reset = true;
        }
        if (connectionToServer != null) {
            int toServerMsgNum;
            int toServerNumBytes;
            connectionToServer.GetStatsIn(out toServerMsgNum, out toServerNumBytes);
            GUILayout.Label("SRV  msg/bytes: " + toServerMsgNum + "/" + toServerNumBytes);
            if (reset) {
                connectionToServer.ResetStats();
            }
        }
        if (connectionToClient != null) {
            int toClientMsgNum;
            int toClientNumBytes;
            connectionToClient.GetStatsIn(out toClientMsgNum, out toClientNumBytes);
            GUILayout.Label("CLI msg/bytes: " + toClientMsgNum + "/" + toClientNumBytes);
            if (reset) {
                connectionToClient.ResetStats();
            }
        }
    }

    private IEnumerator TimeUp()
    {
        if (isServer) {
            NotifyTimeUp();
            yield return new WaitForSeconds(5f);
            var lobbyMgr = PainterLobbyManager.singleton as PainterLobbyUI;
            lobbyMgr.ServerReturnToLobby();
        }
    }

    private void NotifyTimeUp()
    {
        foreach (var ply in FindObjectsOfType<PlayerCharController>()) {
            ply.RpcStopPlaying();
        }
    }
}

public class GameTimeClientMessage : MessageBase
{
    public float sendTime;
}

public class GameTimeServerMessage : MessageBase
{
    public float sendTime;
    public float serverTime;
}