﻿using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Assertions;
using UnityEngine.Networking;


/// <summary>
/// プレーヤーキャラクタ処理
/// </summary>
public class PlayerCharController : NetworkBehaviour
{
    [SerializeField] private float initialMoveSpeed = 5f;
    [SerializeField] private float initialPaintScale = 1f;
    [SerializeField] private float rotationSlerp = 0.4f;
    [SerializeField] private float fireSpeed = 5f;
    [SerializeField] private float fireLifetime = 2f;
    [SerializeField] private GameObject firePrefab;
    [SerializeField] private int attackDamage = 50;

    [SerializeField] private KeyCode keyAttack = KeyCode.Z;
    [SerializeField] private KeyCode keyPaint = KeyCode.X;

    public static PlayerCharController ownerPlayerInstance { get; private set; }
    public PaintedState myPaintState { get; private set; }

    private CharacterAnimatorController charAnimatorContoller;
    private UnityEngine.AI.NavMeshAgent agent;
    private bool playable;
    private List<Skill> skills = new List<Skill>();

    [SyncVar] public string playerName = "player";
    [SyncVar] public float skillRunSpeedScale = 1f;
    [SyncVar] public float skillPaintScale = 1f;
    [SyncVar] public int colorNumber = -1;

    private float runningSpeed { get { return skillRunSpeedScale * initialMoveSpeed; } }

    void Awake()
    {
        agent = GetComponent<UnityEngine.AI.NavMeshAgent>();
        charAnimatorContoller = GetComponent<CharacterAnimatorController>();
        charAnimatorContoller.enabled = false;
    }

    public override void OnStartServer()
    {
        skills.Add(new Skill() { skillType = SkillType.PaintScale, skillValue = 1f });
        skills.Add(new Skill() { skillType = SkillType.RunningSpeed, skillValue = 1f });
        skills.Add(new Skill() { skillType = SkillType.RunningSpeedOnMyPaint, skillValue = 0.2f });
        skills.Add(new Skill() { skillType = SkillType.RunningSpeedOnOtherPaint, skillValue = -0.2f });
        skills.Add(new Skill() { skillType = SkillType.AttackDamage, skillValue = 1f });
    }

    public override void OnStartClient()
    {
    }

    public override void OnStartLocalPlayer()
    {
        playable = true;
        FindObjectsOfType(typeof(NetworkManagerHUD)).Cast<NetworkManagerHUD>().ToList().ForEach(h => h.enabled = false);
        charAnimatorContoller.enableActionTransition = true;
    }

    [ClientRpc]
    public void RpcStopPlaying()
    {
        playable = false;
        charAnimatorContoller.Reset();
    }

    void Update()
    {
        if (isServer) {
            if (NetworkPaintEnv.instance != null) {
                myPaintState = NetworkPaintEnv.instance.GetPaintState(colorNumber, transform.position);
            }
            UpdateSkills();
        }
        if (isLocalPlayer) {
            ownerPlayerInstance = this;
            if (playable && GameController.isInPlay) {
                ControlMyCharacter();
            }
        }
        if (Hud.instance != null) {
            Hud.instance.ShowOccupationGauge(colorNumber);
        }
    }

    private void ControlMyCharacter()
    {
        if (!charAnimatorContoller.IsState(GameAnimType.BlockLocomotion)) {
            if (Input.GetKeyDown(keyPaint)) {
                CmdFire(GameController.instance.worldClock, transform.position, transform.rotation.eulerAngles.y);
            }
            if (Input.GetKeyDown(keyAttack)) {
                CmdAttack();
                charAnimatorContoller.Attack();
            } else {
                var x = Input.GetAxis("Horizontal");
                var z = Input.GetAxis("Vertical");
                var d = new Vector3(x, 0f, z);
                Move(d);
            }
        }
    }

    public void Move(Vector3 direction)
    {
        Debug.Log("move : " + direction.magnitude);
        if (direction.magnitude <= 0.1f) {
            charAnimatorContoller.SetLocomotion(false);
            return;
        }
        direction = Camera.main.transform.TransformVector(direction);
        direction.y = 0f;
        agent.transform.rotation = Quaternion.Slerp(transform.rotation, Quaternion.LookRotation(direction.normalized, Vector3.up), rotationSlerp);
        agent.velocity = transform.forward * runningSpeed;
        charAnimatorContoller.SetLocomotion(true);
    }

    [Command]
    public void CmdFire(float shootTime, Vector3 shootPosition, float shootYangle)
    {
        var forward = Quaternion.Euler(0f, shootYangle, 0f) * Vector3.forward;
        var bulletObj = (GameObject) Instantiate(firePrefab);
        var bullet = bulletObj.GetComponent<Bullet>();
        bullet.shooterNumber = colorNumber;
        bullet.paintScale = skillPaintScale * initialPaintScale;
        bullet.shootClock = shootTime;
        bullet.initialPosition = shootPosition + forward * 0.2f + Vector3.up * 1f;
        bullet.initialVelocity = transform.forward * fireSpeed + transform.up * 2f;
        NetworkServer.Spawn(bulletObj);
        Destroy(bulletObj, fireLifetime);
    }

    [Command]
    public void CmdAttack()
    {
        charAnimatorContoller.Attack();
    }

    /// <summary>
    /// 体についているコリジョン（キックの攻撃判定）の処理。
    /// </summary>
    void OnTriggerEnter(Collider other)
    {
        if (isServer == false) {
            return;
        }
        var combat = other.gameObject.GetComponent<Combat>();
        if (combat != null) {
            var dmg = (int) (attackDamage * GetSkillParam(SkillType.AttackDamage));
            combat.AttackHitByPlayer(dmg);
        }
    }

    private void UpdateSkills()
    {
        skills.ForEach(s => s.Update(this));
        skills.RemoveAll(s => !s.isAlive);
        skillRunSpeedScale = GetSkillParam(SkillType.RunningSpeed) + GetSkillParam(SkillType.RunningSpeedOnMyPaint) + GetSkillParam(SkillType.RunningSpeedOnOtherPaint);
        skillPaintScale = GetSkillParam(SkillType.PaintScale);
    }

    public void AddSkill(Skill skill)
    {
        Assert.IsNotNull(skill);
        skills.Add(skill);
    }
    private float GetSkillParam(SkillType skillType)
    {
        return Mathf.Max(0f, skills.Where(s => s.skillType == skillType).Sum(s => s.effectValue));
    }

    private void Log(string message, bool noFrameCount = false)
    {
        var frameCount = Time.frameCount.ToString() + " ";
        if (noFrameCount) {
            frameCount = "";
        }
        Debug.Log(frameCount + message
                    + " isServer:" + isServer + " isClient:" + isClient + " isLocalPlayer:" + isLocalPlayer + " hasAuthority:" + hasAuthority
                    + " netid:" + netId.Value
                    + " playerControllerId:" + playerControllerId
                    + " instanceId=" + gameObject.GetInstanceID(),
                    gameObject);
    }
}
