﻿using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Networking;
using UnityEngine.Networking.Match;
using System.Linq;
using UnityEngine.SceneManagement;
using System.Collections;

/// <summary>
/// ロビー処理
/// </summary>
public class PainterLobbyManager : NetworkLobbyManager
{
    [SerializeField] private Transform[] lobbyPlayerPositions;
    [SerializeField] private Transform[] gamePlayerPrefabs;
    [SerializeField] private bool debugLog = false;

    public MatchInfo roomHostInfo { get; private set; }
    public MatchInfo joinedInfo { get; private set; }
    public string createdRoomName { get; private set; }
    public MatchInfoSnapshot selectMatchInfo { get; private set; }

    private PainterLocalDiscovery localDiscovery
    {
        get { return gameObject.GetComponent<PainterLocalDiscovery>(); }
    }

    void Start()
    {
        // lobbyPlayerPositions.ToList().ForEach(t => t.gameObject.SetActive(false));
    }
    void OnEnable()
    {
        localDiscovery.onReceieveBroadcast.AddListener(OnReceiveDiscoveryBroadcast);
    }

    void OnDisable()
    {
        localDiscovery.onReceieveBroadcast.RemoveListener(OnReceiveDiscoveryBroadcast);
    }

    private void OnReceiveDiscoveryBroadcast(string address, string data)
    {
        networkAddress = address;
    }

    internal Transform GetPlayerPosition(int slot)
    {
        return lobbyPlayerPositions[slot];
    }

    private void EnablePlayerPositions(bool b)
    {
        lobbyPlayerPositions.ToList().ForEach(t => t.gameObject.SetActive(b));
    }

    protected virtual void OnError(string errorMessage)
    {
        Debug.LogError("OnError: " + errorMessage);
    }

    // ---------------------------- ロビー server ----------------------------------

    #region ロビー server
    public override void OnLobbyServerConnect(NetworkConnection conn)
    {
        base.OnLobbyServerConnect(conn);
        Info("OnLobbyServerConnect: conn=" + conn.hostId + "/" + conn.connectionId);
    }
    public override GameObject OnLobbyServerCreateGamePlayer(NetworkConnection conn, short playerControllerId)
    {
        Info("OnLobbyServerCreateGamePlayer");
        var lp = conn.playerControllers.FirstOrDefault(p => p.playerControllerId == playerControllerId).unetView.GetComponent<PainterLobbyPlayer>();
        var prefab = gamePlayerPrefabs[lp.characterType];
        var spawnPos = GetStartPosition();
        var obj = Instantiate(prefab, spawnPos.position, spawnPos.rotation);
        return obj.gameObject;
    }
    public override GameObject OnLobbyServerCreateLobbyPlayer(NetworkConnection conn, short playerControllerId)
    {
        Info("OnLobbyServerCreateLobbyPlayer conn=" + conn + " playercontrollerId=" + playerControllerId);
        return base.OnLobbyServerCreateLobbyPlayer(conn, playerControllerId);
    }
    public override void OnLobbyServerPlayerRemoved(NetworkConnection conn, short playerControllerId)
    {
        base.OnLobbyServerPlayerRemoved(conn, playerControllerId);
        Info("OnLobbyServerPlayerRemoved! conn=" + conn + " playerControllerId=" + playerControllerId);
    }
    public override void OnLobbyServerDisconnect(NetworkConnection conn)
    {
        base.OnLobbyServerDisconnect(conn);
        Info("OnLobbyServerDisconnect " + conn.hostId + "/" + conn.connectionId);
    }
    public override void OnLobbyServerPlayersReady()
    {
        base.OnLobbyServerPlayersReady();
        Info("OnLobbyServerPlayersReady");
    }
    public override void OnLobbyServerSceneChanged(string sceneName)
    {
        base.OnLobbyServerSceneChanged(sceneName);
        Info("OnLobbyServerSceneChanged " + sceneName);
    }
    public override bool OnLobbyServerSceneLoadedForPlayer(GameObject lobbyPlayerObj, GameObject gamePlayerObj)
    {
        Info("OnLobbyServerSceneLoadedForPlayer lobbyPlayer=" + lobbyPlayerObj + " / gamePlayer=" + gamePlayerObj);
        var lobbyPlayer = lobbyPlayerObj.GetComponent<PainterLobbyPlayer>();
        var playerChar = gamePlayerObj.GetComponent<PlayerCharController>();
        playerChar.colorNumber = lobbyPlayer.slot;
        playerChar.playerName = lobbyPlayer.playerName;
        return true;
    }
    public override void OnLobbyStartServer()
    {
        base.OnLobbyStartServer();
        Info("OnLobbyStartServer");
    }
    public override void OnServerAddPlayer(NetworkConnection conn, short playerControllerId)
    {
        base.OnServerAddPlayer(conn, playerControllerId);
        Info("OnServerAddPlayer " + conn.hostId + "/" + conn.connectionId + " playerControllerId=" + playerControllerId);
    }

    #endregion ロビー server

    // ---------------------------- ロビー client ----------------------------------

    #region ロビー client

    public override void OnLobbyClientAddPlayerFailed()
    {
        base.OnLobbyClientAddPlayerFailed();
    }
    public override void OnLobbyClientConnect(NetworkConnection conn)
    {
        base.OnLobbyClientConnect(conn);
        Info("OnLobbyClientConnect! connection id =" + conn.connectionId);
    }
    public override void OnLobbyClientDisconnect(NetworkConnection conn)
    {
        base.OnLobbyClientDisconnect(conn);
        Info("OnLobbyClientDisconnect! connection id =" + conn.connectionId);
    }
    public override void OnLobbyClientEnter()
    {
        base.OnLobbyClientEnter();
        Info("OnLobbyClientEnter! ");
    }
    public override void OnLobbyClientExit()
    {
        base.OnLobbyClientExit();
        Info("OnLobbyClientExit! ");
    }
    public override void OnLobbyClientSceneChanged(NetworkConnection conn)
    {
        base.OnLobbyClientSceneChanged(conn);
        Info("OnLobbyClientSceneChanged: conn=" + conn.hostId + "/" + conn.connectionId + " scene=" + SceneManager.GetSceneAt(0).name);
        var sceneName = SceneManager.GetSceneAt(0).name;
        EnablePlayerPositions(sceneName != "game");
    }
    public override void OnLobbyStartClient(NetworkClient lobbyClient)
    {
        Info("OnLobbyStartClient client conn=" + lobbyClient);
    }
    public override void OnLobbyStopClient()
    {
        base.OnLobbyStopClient();
        Info("OnLobbyStopClient");
    }

    #endregion ロビー client

    // ---------------------------- ロビー host ----------------------------------

    #region ロビー host
    public override void OnLobbyStartHost()
    {
        Info("OnLobbyStartHost");
    }
    public override void OnLobbyStopHost()
    {
        base.OnLobbyStopHost();
        Info("OnLobbyStopHost");
    }

    #endregion

    // ---------------------------- マッチメイク ----------------------------------

    #region マッチメイク

    public override void OnMatchCreate(bool success, string extendedInfo, MatchInfo matchInfo)
    {
        Info("OnMatchCreate " + success + " /" + extendedInfo + "/" + matchInfo);
        if (success == false) {
            OnError("ルーム作成に失敗しました。ネットワーク状況を確認してください。" + extendedInfo);
        } else {
            roomHostInfo = matchInfo;
        }
    }

    public override void OnMatchJoined(bool success, string extendedInfo, MatchInfo matchInfo)
    {
        Info("OnMatchJoined " + success + " /" + extendedInfo + "/" + matchInfo);
        if (success == false) {
            OnError("ルームjoinに失敗しました");
        } else {
            joinedInfo = matchInfo;
        }
    }
    public override void OnMatchList(bool success, string extendedInfo, List<MatchInfoSnapshot> matchList)
    {
        Info("OnMatchList " + success + " /" + extendedInfo + "\nmatchlist:\n" + string.Join("\n", matchList.Select(m => m.ToString()).ToArray()));
    }
    public override void OnSetMatchAttributes(bool success, string extendedInfo)
    {
        Info("OnMatchCreate " + success + " /" + extendedInfo);
    }
    public override void OnDestroyMatch(bool success, string extendedInfo)
    {
        Info("OnDestroyMatch " + success + "/" + extendedInfo);
    }
    protected void CreateRoom(string roomName)
    {
        roomHostInfo = null;
        matchMaker.CreateMatch(roomName, matchSize, true, "", "", "", 0, requestDomain, OnMatchCreate);
    }

    protected void ListMatches()
    {
        matchMaker.ListMatches(0, 100, "", false, 0, requestDomain, OnMatchList);
    }

    protected void JoinRoom(MatchInfoSnapshot m)
    {
        joinedInfo = null;
        matchMaker.JoinMatch(m.networkId, "", "", "", 0, requestDomain, OnMatchJoined);
    }

    private const int requestDomain = 1;

    #endregion

    // ---------------------------- NetworkManager ----------------------------------

    #region networking server/client 処理

    public override void OnServerConnect(NetworkConnection conn)
    {
        base.OnServerConnect(conn);
        Info("OnServerConnect");
    }
    public override void OnServerDisconnect(NetworkConnection conn)
    {
        base.OnServerDisconnect(conn);
        Info("OnServerDisconnect");
    }
    public override void OnServerSceneChanged(string sceneName)
    {
        base.OnServerSceneChanged(sceneName);
        Info("OnServerSceneChanged " + sceneName);
    }
    public override void OnStartClient(NetworkClient lobbyClient)
    {
        base.OnStartClient(lobbyClient);
        Info("OnStartClient");
    }
    public override void OnStartHost()
    {
        base.OnStartHost();
        Info("OnStartHost");
    }
    public override void OnStartServer()
    {
        base.OnStartServer();
        Info("OnStartServer");
    }
    public override void OnStopClient()
    {
        base.OnStopClient();
        Info("OnStopClient");
    }
    public override void OnStopHost()
    {
        base.OnStopHost();
        Info("OnStopHost");
    }
    public override void OnClientConnect(NetworkConnection conn)
    {
        Info("OnClientConnect");
        base.OnClientConnect(conn);
    }
    public override void OnClientDisconnect(NetworkConnection conn)
    {
        Info("OnClientDisconnect");
        base.OnClientDisconnect(conn);
    }
    public override void OnClientSceneChanged(NetworkConnection conn)
    {
        Info("OnClientSceneChanged");
        base.OnClientSceneChanged(conn);
    }

    #endregion

    // ---------------------------- ext ----------------------------------

    public void InitializeLocalDiscovery()
    {
        localDiscovery.Initialize();
    }

    public IEnumerator CreateRoomAndStartHost(string roomName)
    {
        createdRoomName = roomName;
        StartMatchMaker();
        CreateRoom(roomName);
        yield return new WaitUntil(() => roomHostInfo != null);
        NetworkServer.Listen(roomHostInfo, networkPort);
        StartHost(roomHostInfo);
    }

    public void StartMatchMakerAndRoomSearch()
    {
        StartMatchMaker();
        selectMatchInfo = null;
        ListMatches();
    }

    public void SetJoinRoom(MatchInfoSnapshot selectInfo)
    {
        selectMatchInfo = selectInfo;
    }

    public IEnumerator JoinRoomAndStartClient()
    {
        JoinRoom(selectMatchInfo);
        yield return new WaitUntil(() => joinedInfo != null);
        StartClient(joinedInfo);
    }

    public void StopAll()
    {
        StopLan();
        StopMatchMaker();
    }

    public void StopLan()
    {
        StopHost();
        StopClient();
        if (localDiscovery.isClient || localDiscovery.isServer) {
            localDiscovery.StopBroadcast();
        }
    }

    public bool StartHostDiscovery()
    {
        localDiscovery.Initialize();
        if (localDiscovery.StartAsClient() == false) {
            Debug.LogError("cannot start localDiscovery as client");
            OnError("LAN discovery ソケットを開けません。");
            return false;
        }
        return true;
    }


    protected void Info(string msg, UnityEngine.Object origin = null)
    {
        if (debugLog) {
            Debug.LogWarning("[INFO]:" + Time.frameCount + "/" + Time.realtimeSinceStartup + " " + msg, origin);
        }
    }
}
